#include "MemoryNameSystemComponentPlugin.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>



namespace armarx::armem::client::plugins
{

    MemoryNameSystemComponentPlugin::~MemoryNameSystemComponentPlugin()
    {}


    void MemoryNameSystemComponentPlugin::postCreatePropertyDefinitions(armarx::PropertyDefinitionsPtr& properties)
    {
        if (!properties->hasDefinition(makePropertyName(PROPERTY_MNS_NAME_NAME)))
        {
            properties->defineOptionalProperty<std::string>(
                makePropertyName(PROPERTY_MNS_NAME_NAME),
                parent<MemoryNameSystemComponentPluginUser>().memoryNameSystemName,
                "Name of the Memory Name System (MNS) component.");
        }
        if (!properties->hasDefinition(makePropertyName(PROPERTY_MNS_ENABLED_NAME)))
        {
            properties->defineOptionalProperty<bool>(
                makePropertyName(PROPERTY_MNS_ENABLED_NAME),
                parent<MemoryNameSystemComponentPluginUser>().memoryNameSystemEnabled,
                "Whether to use (and depend on) the Memory Name System (MNS)."
                "\nSet to false to use this memory as a stand-alone.");
        }

        // Subscribe topics by single servers, use this as a prefix.
        properties->topic<MemoryListenerInterface>("MemoryUpdates");
    }

    void MemoryNameSystemComponentPlugin::preOnInitComponent()
    {
        if (isMemoryNameSystemEnabled())
        {
            parent().usingProxy(getMemoryNameSystemName());
        }
    }

    void MemoryNameSystemComponentPlugin::preOnConnectComponent()
    {
        if (isMemoryNameSystemEnabled())
        {
            ARMARX_DEBUG << "Creating MemoryNameSystem client with owning component '" << parent().getName() << "'.";
            parent<MemoryNameSystemComponentPluginUser>().memoryNameSystem =
                MemoryNameSystem(getMemoryNameSystem(), &parent());
        }
    }

    bool MemoryNameSystemComponentPlugin::isMemoryNameSystemEnabled()
    {
        return parentDerives<Component>() ?
               parent<Component>().getProperty<bool>(makePropertyName(PROPERTY_MNS_ENABLED_NAME)) :
               parent<MemoryNameSystemComponentPluginUser>().memoryNameSystemEnabled;
    }

    std::string MemoryNameSystemComponentPlugin::getMemoryNameSystemName()
    {
        return parentDerives<Component>() ?
               parent<Component>().getProperty<std::string>(makePropertyName(PROPERTY_MNS_NAME_NAME)) :
               std::string{parent<MemoryNameSystemComponentPluginUser>().memoryNameSystemName};
    }

    mns::MemoryNameSystemInterfacePrx MemoryNameSystemComponentPlugin::getMemoryNameSystem()
    {
        return isMemoryNameSystemEnabled() && parentDerives<Component>()
               ? parent<Component>().getProxy<mns::MemoryNameSystemInterfacePrx>(getMemoryNameSystemName())
               : nullptr;
    }

}

namespace armarx::armem::client
{

    MemoryNameSystemComponentPluginUser::MemoryNameSystemComponentPluginUser()
    {
        addPlugin(plugin);
    }

    MemoryNameSystemComponentPluginUser::~MemoryNameSystemComponentPluginUser()
    {
    }

    void MemoryNameSystemComponentPluginUser::memoryUpdated(
        const std::vector<data::MemoryID>& updatedSnapshotIDs, const Ice::Current&)
    {
        memoryNameSystem.updated(updatedSnapshotIDs);
    }

    armem::data::WaitForMemoryResult MemoryNameSystemComponentPluginUser::useMemoryServer(const MemoryID& id)
    {
        armem::data::WaitForMemoryResult result;
        try
        {
            // Add dependency.
            result.proxy = memoryNameSystem.useServer(id);
            result.success = true;
        }
        catch (const armem::error::CouldNotResolveMemoryServer& e)
        {
            result.success = false;
            result.errorMessage = e.what();
        }
        return result;
    }

    armem::data::WaitForMemoryResult MemoryNameSystemComponentPluginUser::useMemoryServer(const std::string& memoryName)
    {
        return useMemoryServer(MemoryID().withMemoryName(memoryName));
    }

    armem::data::WaitForMemoryResult MemoryNameSystemComponentPluginUser::waitForMemoryServer(const std::string& memoryName)
    {
        armem::data::WaitForMemoryResult result;
        try
        {
            result.proxy = memoryNameSystem.waitForServer(MemoryID().withMemoryName(memoryName));
            result.success = true;
        }
        catch (const armem::error::CouldNotResolveMemoryServer& e)
        {
            result.success = false;
            result.errorMessage = e.what();
        }
        return result;
    }

    armem::data::ResolveMemoryNameResult MemoryNameSystemComponentPluginUser::resolveMemoryServer(const std::string& memoryName)
    {
        armem::data::ResolveMemoryNameResult result;
        try
        {
            result.proxy = memoryNameSystem.resolveServer(MemoryID().withMemoryName(memoryName));
            result.success = true;
        }
        catch (const armem::error::CouldNotResolveMemoryServer& e)
        {
            result.success = false;
            result.errorMessage = e.what();
        }
        return result;
    }

}

