/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Object.h"

// ArmarX
#include "SimoxUtility/algorithm/get_map_keys_values.h"


namespace armarx::aron::typenavigator
{

    // constructors
    ObjectNavigator::ObjectNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path)
    {
    }

    ObjectNavigator::ObjectNavigator(const std::string& name, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path)
    {
        setObjectName(name);
    }

    ObjectNavigator::ObjectNavigator(const type::AronObjectPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path),
        detail::NavigatorBase<type::AronObject, ObjectNavigator>(o)
    {
        for (const auto& [key, t] : o->elementTypes)
        {
            memberTypes[key] = FromAronType(t);
        }
    }

    // operators
    bool ObjectNavigator::operator==(const ObjectNavigator& other) const
    {
        if (getMaybe() != other.getMaybe())
        {
            return false;
        }

        return getObjectName() == other.getObjectName();
    }

    bool ObjectNavigator::checkObjectName(const std::string& s) const
    {
        if (s.empty())
        {
            throw error::AronException("ObjectNavigator", "getResult", "The object name is empty.", getPath());
        }
        return true;
    }

    // public member functions
    std::map<std::string, NavigatorPtr> ObjectNavigator::getMemberTypes() const
    {
        return memberTypes;
    }

    NavigatorPtr ObjectNavigator::getMemberType(const std::string& s) const
    {
        if (memberTypes.find(s) == memberTypes.end())
        {
            throw error::StringNotValidException("ObjectNavigator", "getMemberType", "Member not set. The list of all members is: " + simox::alg::to_string(simox::alg::get_keys(memberTypes)), s);
        }
        return memberTypes.at(s);
    }

    void ObjectNavigator::addMemberType(const std::string& k, const NavigatorPtr& v)
    {
        if (k.empty())
        {
            throw error::AronException("ObjectNavigator", "addAcceptedType", "Cannot set an element with an empty key.", getPath());
        }

        ARMARX_CHECK_NOT_NULL(v);

        this->aron->elementTypes[k] = v->toAronTypePtr();
        memberTypes[k] = v;
    }

    void ObjectNavigator::setObjectName(const std::string& n)
    {
        //path.setRootIdentifier(n);
        this->aron->objectName = n;
    }

    void ObjectNavigator::setExtends(const ObjectNavigatorPtr& p)
    {
        ARMARX_CHECK_NOT_NULL(p);
        type::AronObjectPtr ex = p->toAronObjectPtr();
        ARMARX_CHECK_NOT_NULL(ex);
        extends = p;
    }

    bool ObjectNavigator::hasMemberType(const std::string& k) const
    {
        return memberTypes.count(k) > 0;
    }

    std::string ObjectNavigator::getObjectName() const
    {
        return this->aron->objectName;
    }

    ObjectNavigatorPtr ObjectNavigator::getExtends() const
    {
        return extends;
    }

    std::vector<std::string> ObjectNavigator::getAllKeys() const
    {
        std::vector<std::string> ret;
        for (const auto& [k, _] : memberTypes)
        {
            ret.push_back(k);
        }
        return ret;
    }

    type::AronObjectPtr ObjectNavigator::toAronObjectPtr() const
    {
        // TODO: Shall we allow empty objects?
        //if(acceptedTypeNavigators.empty())
        //{
        //    throw exception::AronExceptionWithPathInfo("ObjectNavigator", "getResult", "No accepted types set", getPath());
        //}
        return this->aron;
    }

    // virtual implementations
    std::vector<NavigatorPtr> ObjectNavigator::getChildren() const
    {
        std::vector<NavigatorPtr> ret;
        for (const auto& [k, t] : memberTypes)
        {
            ret.push_back(t);
        }
        return ret;
    }

    size_t ObjectNavigator::childrenSize() const
    {
        return memberTypes.size();
    }

    std::string ObjectNavigator::getName() const
    {
        return "AronObjectType<" + this->aron->objectName + ">";
    }
}

